@import url("theme.css");

.highlight a {
  text-decoration: underline;
}

.deprecated p {
  padding: 10px 7px 10px 10px;
  color: #b94a48;
  background-color: #f3e5e5;
  border: 1px solid #eed3d7;
}

.deprecated p span.versionmodified {
  font-weight: bold;
}
